#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbdrv/usbdrv.h"

enum CLOCKMUX {
  MUX_XO, MUX_SOFT, MUX_XO_4, MUX_XO_2, MUX_XO_8, MUX_XO_16, MUX_XO_32, MUX_XO_64
};

/* declared volatile so we can manipulate them using gdb */
volatile unsigned int fill_pattern = 0;
volatile int fill_count = 1;
volatile int do_fill = 0;

volatile unsigned int clock_source = MUX_XO;
volatile unsigned int trigger_config = 0;
volatile unsigned int trigger_delay = 0;
volatile unsigned long sample_frequency = 0;

static void seed_on()  { PORTB |= 1 << 5;}
static void seed_off() { PORTB &= ~(1 << 5);}
static void gate_on()  { PORTD &= ~(1 << 5);}
static void gate_off() { PORTD |= 1 << 5;}
static void we_on()    { PORTB &= ~(1 << 2);}
static void we_off()   { PORTB |= 1 << 2;}
static void oe_on()    { PORTB &= ~(1 << 0);}
static void oe_off()   { PORTB |= 1 << 0;}

static void mux(int source)
{
  PORTB = (PORTB & ~0x18) | ((source & 0x3) << 3);
  PORTD = (PORTD & ~0x2) | ((source & 0x4) >> 1);
}

/* toggle clock pin */
static void toggle()
{
  PORTB ^= 0x2;
}

static void fgen_on()
{
  TCCR1A |= 1 << COM1A0;
}

__attribute__((always_inline))
static void fgen_off()
{
  TCCR1A &= ~(1 << COM1A0);
}

void bus_output()
{
  oe_off();
  gate_off();
  DDRC |= 0x3f;
  DDRD |= 0xc0;
}

void bus_tristate()
{
  DDRC &= ~0x3f;
  DDRD &= ~0xc0;
  PORTC &= ~0x3f;
  PORTD &= ~0xc0;
}

void prepwrite()
{
  fgen_off();
  mux(MUX_SOFT);
  gate_off();
  oe_off();
  bus_output();
  we_on();
}

void prepread()
{
  fgen_off();
  mux(MUX_SOFT);
  we_off();
  gate_off();
  bus_tristate();
  oe_on();
}

void prepgate()
{
  fgen_off();
  mux(MUX_SOFT);
  we_off();
  oe_off();
  bus_tristate();
  gate_on();
}

void preprecord()
{
  mux(clock_source);
  if (clock_source == MUX_SOFT)
    fgen_on();
  else
    fgen_off();

  oe_off();
  bus_tristate();
  gate_on();
  we_on();
}

static void bus_out(unsigned int value)
{
  PORTC = (PORTC & ~0x3f) | (value & 0x3f);
  PORTD = (PORTD & ~0xc0) | (value & 0xc0);
}

static void bus_write(unsigned int value)
{
  bus_out(value);
  toggle();
}

unsigned char bus_in()
{
  return (PINC & 0x3f) << 2 | ((PIND & 0xc0) >> 6);
}

unsigned char bus_read()
{
  toggle();
  return bus_in();
}

static void
fill()
{
  unsigned int i = 0;
  prepwrite();
  for (i=0; i<65534; i++) {
    bus_write(fill_count ? i : fill_pattern);
  }
  bus_tristate();
}

__attribute__((constructor))
void reset_hw()
{
  gate_off();
  oe_off();
  we_off();
  mux(MUX_SOFT);
  /* MUX  */
  DDRB |= (1 << 3 | 1 << 4);
  DDRD |= 1 << 1;
  /*   GATE */
  DDRD |= 1 << 5;
  /*   WE */
  DDRB |= 1 << 2;
  /*   OE */
  DDRB |= 1 << 0;
  /* softclock */
  DDRB |= 1 << 1;
  /* lfsr_init */
  DDRB |= 1 << 5;
  /* frequency generator clock source, 8MHz default */
  TCCR1B = (TCCR1B & ~0x7) | 1;
  TCCR1B |= 1 << WGM12;
  OCR1AH = 0;
  OCR1AL = 0;
  seed_on();
  toggle();
  toggle();
  seed_off();
  TIMSK |= 1 << OCIE2;
}

enum {
  USBRQ_SET_CLOCK_SRC = 0,  /* clock multiplexer setting for recording */
  USBRQ_RESERVED1 = 1,
  USBRQ_SET_SOFTCLOCK = 2,  /* value = timer1 prescaler, index = OCR (16-bit) */
  USBRQ_RESERVED2 = 3,
  USBRQ_READ_MEM = 4,       /* read bytes from Cache chips */
  USBRQ_FILL_MEM = 5,       /* value: fill with address, index:
			       pattern used otherise */
  USBRQ_RECORD = 6,         /* start sampling (free running) */
  USBRQ_TRIGGER_CONFIG = 7, /* set ISC11,ISC10 in MCUCR, bit 3
			       enables internal pullup, bit 4 USB
			       lockout for improved timing (TODO) */
  USBRQ_ARM = 8,            /* start sampling and arm trigger */
  USBRQ_SNAPSHOT = 9,       /* look at the inputs using µC */
  USBRQ_SAMPLE_OSC = 10,    /* return oscillator frequency measured
			       at startup (unsigned long) */
  USBRQ_RESET = 11,         /* reset the hardware */
  USBRQ_TRIGGER_DELAY = 12, /* word: timer2 clock source, index: OCR2 */
  USBRQ_STATUS = 13,        /* return status word ...,GATE,OE,WE */
};

usbMsgLen_t usbFunctionSetup(uchar setupData[8])
{
    usbRequest_t *rq = (void *)setupData;
    static usbWord_t data;
    switch(rq->bRequest){
    case USBRQ_SET_CLOCK_SRC:
      clock_source = rq->wValue.word;
      break;
    case USBRQ_SET_SOFTCLOCK:
      TCCR1B = (TCCR1B & ~0x7 ) | (rq->wValue.word & 0x7); /* clocksource */
      OCR1AH = rq->wIndex.bytes[1];
      OCR1AL = rq->wIndex.bytes[0];
      break;
    case USBRQ_READ_MEM:
      prepread();
      return USB_NO_MSG; /* use usbFunctionRead() */
    case USBRQ_FILL_MEM:
      fill_count = rq->wValue.word;
      fill_pattern = rq->wIndex.word;
      do_fill++;
      break;
    case USBRQ_RECORD:
      preprecord();
      break;
    case USBRQ_ARM:
      preprecord();
      GICR |= (1<<INT1);
      break;
    case USBRQ_SNAPSHOT:
      prepgate();
      data.bytes[0] = bus_in();
      usbMsgPtr = (uchar *)&data;
      return sizeof(data.bytes[0]);
      break;
    case USBRQ_TRIGGER_CONFIG:
      MCUCR = (MCUCR & ~(3 << 2)) | ((rq->wValue.bytes[0] & 3) << 2);
      PORTD = (PORTD & ~(1 << 3)) | ((rq->wValue.bytes[0] & 4) << 1); 
     break;
    case USBRQ_SAMPLE_OSC:
      usbMsgPtr = (uchar *)&sample_frequency;
      return sizeof(sample_frequency);
    case USBRQ_RESET:
      reset_hw();
      break;
    case USBRQ_TRIGGER_DELAY:
      trigger_delay = rq->wValue.word;
      OCR2 = rq->wIndex.word;
      break;
    case USBRQ_STATUS:
      data.word = 0;
      if (!(PORTB & (1 << 2)))
	data.word |= 1 << 0;
      if (!(PORTB & (1 << 0)))
	data.word |= 1 << 1;
      if (!(PORTD & (1 << 5)))
	data.word |= 1 << 2;
      usbMsgPtr = (uchar *)&data;
      return sizeof(data);
      break;
    default:
      break;
    }
    return 0;
}

uchar usbFunctionRead(uchar *data, uchar len)
{
    uchar i;
    for(i = 0; i < len;)
      data[i++] = bus_read();
    return len;
}

static void
measure_sample_frequency()
{
  prepwrite();
  /* clear buffer */
  mux(MUX_XO);
  bus_out(0);
  _delay_ms(10);
  /* write a 200 us pulse */
  bus_out(0xff);
  _delay_us(40);
  _delay_us(40);
  _delay_us(40);
  _delay_us(40);
  _delay_us(40);
  we_off();
  prepread();
  /* find the pulse */
  while(!bus_read());
  sample_frequency = 5000;
  /* count samples */
  while(bus_read())
    sample_frequency += 5000;
  return;
}

int
main()
{
  unsigned char i;
  usbDeviceDisconnect();

  measure_sample_frequency();

  for(i=0;--i;){             /* fake USB disconnect for > 250 ms */
    wdt_reset();
    _delay_ms(1);
  }

  usbDeviceConnect();
  usbInit();
  sei();

  while (1) {
    usbPoll();
    if (do_fill)
      fill(), do_fill--;
  }
}

/* ext. trigger */
ISR(INT1_vect)
{
  if(!trigger_delay)
    {
      mux(MUX_SOFT);
      fgen_off();
      we_off();
    }
  else
    {
      TCNT2 = 0;
      TCCR2 |= trigger_delay;
    }
  GICR &= ~(1<<INT1);
}

/* timer to end recording */
ISR(TIMER2_COMP_vect) {
  mux(MUX_SOFT);
  fgen_off();
  we_off();
  TCCR2 &= ~7;
}
